@@ -32,11 +32,21 @@ class PiusSigner(object):
   GPG_GOOD_PASS = '[GNUPG:] GOOD_PASSPHRASE'
   GPG_SIG_BEG = '[GNUPG:] BEGIN_SIGNING'
   GPG_SIG_CREATED = '[GNUPG:] SIG_CREATED'
+  GPG_PROGRESS = '[GNUPG:] PROGRESS'
+  GPG_PINENTRY_LAUNCHED = '[GNUPG:] PINENTRY_LAUNCHED'
+  GPG_KEY_CONSIDERED = '[GNUPG:] KEY_CONSIDERED'
 
-  def __init__(self, signer, mode, keyring, gpg_path, tmpdir, outdir,
-               encrypt_outfiles, mail, mailer, verbose, sort_keyring,
+  def __init__(self, signer, force_signer, mode, keyring, gpg_path, tmpdir,
+               outdir, encrypt_outfiles, mail, mailer, verbose, sort_keyring,
                policy_url, mail_host):
     self.signer = signer
+    if not force_signer:
+      # If force_signer is not specified let gpg guess by using the main keyid
+      self.force_signer = self.signer
+    else:
+      # If force_signer is specified make sure that gpg uses this keyid by
+      # putting '!' at the end
+      self.force_signer = force_signer + '!'
     self.mode = mode
     self.keyring = keyring
     self.sort_keyring = sort_keyring
@@ -423,6 +433,12 @@ class PiusSigner(object):
         # we get a ENC_INV.
         debug('Got GPG_KEY_EXP')
         continue
+      elif PiusSigner.GPG_KEY_CONSIDERED in line:
+        debug('Got KEY_CONSIDERED')
+        continue
+      elif PiusSigner.GPG_PROGRESS in line:
+        debug('Got skippable stuff')
+        continue
       else:
         raise EncryptionUnknownError(line)
 
@@ -545,7 +561,10 @@ class PiusSigner(object):
     line = ''
     while line not in (string,):
       debug('Waiting for line %s' % string)
-      line = fd.readline().strip()
+      raw_line = fd.readline()
+      if not raw_line:
+        raise GpgUnknownError('gpg output unexpectedly ended')
+      line = raw_line.strip()
       debug('got line %s' % line)
 
   def sign_uid(self, key, index, level):
@@ -559,12 +578,12 @@ class PiusSigner(object):
     # Note that if passphrase-fd is different from command-fd, nothing works.
     cmd = [self.gpg] + self.gpg_base_opts + self.gpg_quiet_opts + \
       self.gpg_fd_opts + keyring + [
-          '-u', self.signer,
-      ] + agent + [
+          '-u', self.force_signer,
+      ] + agent + self.policy_opts() + [
           '--default-cert-level', level,
           '--no-ask-cert-level',
           '--edit-key', key,
-      ] + self.policy_opts()
+      ]  # NB: keep the `--edit-key <key>` at the very end of this list!
     logcmd(cmd)
     gpg = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
@@ -603,6 +622,9 @@ class PiusSigner(object):
         print '  UID already signed'
         gpg.stdin.write('quit\n')
         return False
+      elif PiusSigner.GPG_KEY_CONSIDERED in line:
+        debug('Got KEY_CONSIDERED')
+        continue
       elif (PiusSigner.GPG_KEY_EXP in line or
             PiusSigner.GPG_SIG_EXP in line):
         # The user has an expired signing or encryption key, keep going
@@ -698,7 +720,7 @@ class PiusSigner(object):
       print '  Signed encrypted keys: '
       for uid in uids:
         if uid['status'] != 'r' and uid['result']:
-          print '    %(id)s: %(enc_file)s' % uids
+          print '    %(id)s: %(enc_file)s' % uid
 
   def sign_all_uids(self, key, level):
     '''The main function that signs all the UIDs on a given key.'''
@@ -823,7 +845,7 @@ class PiusSigner(object):
           '--keyring', self.tmp_keyring,
           '--no-options',
           '--always-trust',
-          '-u', self.signer,
+          '-u', self.force_signer,
           '-aes',
           '-r', keyid,
           '-r', self.signer,
@@ -858,6 +880,9 @@ class PiusSigner(object):
       elif PiusSigner.GPG_ENC_INV in line:
         debug('Got GPG_ENC_INV')
         raise EncryptionKeyError
+      elif PiusSigner.GPG_KEY_CONSIDERED in line:
+        debug('Got KEY_CONSIDERED')
+        continue
       elif (PiusSigner.GPG_KEY_EXP in line or
             PiusSigner.GPG_SIG_EXP in line):
         # These just mean we passed a given key/sig that's expired, there
@@ -869,7 +894,9 @@ class PiusSigner(object):
             PiusSigner.GPG_NEED_PASS in line or
             PiusSigner.GPG_GOOD_PASS in line or
             PiusSigner.GPG_SIG_BEG in line or
-            PiusSigner.GPG_SIG_CREATED in line):
+            PiusSigner.GPG_SIG_CREATED in line or
+            PiusSigner.GPG_PROGRESS in line or
+            PiusSigner.GPG_PINENTRY_LAUNCHED in line):
         debug('Got skippable stuff')
         continue
       else:
